import type {IBaseTroonia, TrooniaState} from '../common/troonia';
import {io, Socket} from 'socket.io-client';
import {World} from './world';
import {get_viewport_size} from './util';
import {PerspectiveCamera, WebGLRenderer} from 'three';
import {GUI} from 'dat.gui';
import {OrbitControls} from 'three/examples/jsm/controls/OrbitControls';
import {BaseTroonia} from '../common/troonia';
import {TestScene} from '../common/scenes';
import type {IPlayerData} from '../common/player';


export class Troonia extends BaseTroonia implements IBaseTroonia {
  socket: Socket;
  renderer: WebGLRenderer;
  camera: PerspectiveCamera;
  gui: GUI = new GUI();
  world: World;

  constructor() {
    super();
    this.socket = io();

    const viewport = get_viewport_size();

    this.renderer = new WebGLRenderer({antialias: true});
    this.renderer.setSize(viewport.width, viewport.height);
    document.body.appendChild(this.renderer.domElement);

    this.camera =
        new PerspectiveCamera(75, viewport.width / viewport.height, 0.1, 2000);
    this.camera.position.set(0, 50, 0);
    this.camera.lookAt(0, 0, 0);

    const controls = new OrbitControls(this.camera, this.renderer.domElement);
    controls.target.set(0, 1.25, 0);
    controls.update();

    this.socket.on('connect', () => {
      this.onConnect();
    });
    this.socket.on('disconnect', (message: any) => {
      this.onDisconnect(message);
    });

    this.world = new World(this);
    this.world.load(TestScene);

    this.socket.on('syncState', (state: TrooniaState) => {
      this.world.physics.bodies = state.objects;
      this.world.physics.syncState();
    });

    this.socket.on('syncPlayers', (players: {[id: string]: IPlayerData}) => {
      for (const player in players) {
        if (player !== this.socket.id) {
          this.world.addPlayer(player);
        }
      }
    });

    this.socket.on('addPlayer', (id: string) => {
      if (id !== this.socket.id) {
        this.world.addPlayer(id);
      }
    });

    this.socket.on('removePlayer', (id: string) => {
      if (id !== this.socket.id) {
        this.world.removePlayer(id);
      }
    });

    document.addEventListener('keydown', (event: KeyboardEvent) => {
      switch (event.code) {
        case 'KeyW':
          this.world.player.vec3.z = 1;
          break;
        case 'KeyA':
          this.world.player.vec3.x = 1;
          break;
        case 'Space':
          this.world.player.vec3.y = 1;
          break;
        case 'KeyS':
          this.world.player.vec3.z = -1;
          break;
        case 'KeyD':
          this.world.player.vec3.x = -1;
          break;
      }
      this.socket.emit('updatePlayer', this.world.player);
    }, false);
    document.addEventListener('keyup', (event: KeyboardEvent) => {
      switch (event.code) {
        case 'KeyW':
        case 'KeyS':
          this.world.player.vec3.z = 0;
          break;
        case 'KeyA':
        case 'KeyD':
          this.world.player.vec3.x = 0;
          break;
        case 'Space':
          this.world.player.vec3.y = 0;
          break;
      }
      this.socket.emit('updatePlayer', this.world.player);
    }, false);
  }

  onConnect() {
    console.log('Connected to Troonia.');
    this.world.addPlayer();
  }

  onDisconnect(message: any) {
    this.world.removePlayer();
    console.log(`Disconnected from Troonia: ${message}.`);
  }

  update() {
    const delta = this.clock.getDelta();
    this.world.update(delta);
    this.renderer.render(this.world.scene, this.camera);
  }
}